import jax_dataclasses as jdc

from abc import ABC, abstractmethod

from quantumspectra_2024.common.absorption.AbsorptionSpectrum import (
    AbsorptionSpectrum,
)


@jdc.pytree_dataclass(kw_only=True)
class AbsorptionModel(ABC):
    """Represents a model for generating absorption spectra.

    All models include a `get_absorption` method that returns an `AbsorptionSpectrum` object.
    An `apply_electric_field` method is also included to apply an electric field to base models for Stark effect.

    Parameters
    ----------
    start_energy : float
        absorption spectrum's starting energy (wavenumbers).
    end_energy : float
        absorption spectrum's ending energy (wavenumbers).
    num_points : int
        absorption spectrum's number of points (unitless).
    """

    #: absorption spectrum's starting energy (wavenumbers).
    start_energy: float = 0.0
    #: absorption spectrum's ending energy (wavenumbers).
    end_energy: float = 20_000.0
    #: absorption spectrum's number of points (unitless).
    num_points: int = 2_001

    @abstractmethod
    def get_absorption(self) -> AbsorptionSpectrum:
        """Compute the absorption spectrum for the model.

        Returns
        -------
        AbsorptionSpectrum
            the model's parameterized absorption spectrum.
        """
        raise NotImplementedError

    @abstractmethod
    def apply_electric_field(
        field_strength: float,
        field_delta_dipole: float,
        field_delta_polarizability: float,
    ) -> "AbsorptionModel":
        """Applies an electric field to the model. Returns a new instance of the model.

        Parameters
        ----------
        field_strength : float
            the strength of the electric field.
        field_delta_dipole : float
            the change in dipole moment due to the electric field.
        field_delta_polarizability : float
            the change in polarizability due to the electric field.

        Returns
        -------
        AbsorptionModel
            the model with the electric field applied.
        """
        raise NotImplementedError



import numpy as np
import matplotlib.pyplot as plt

import jax_dataclasses as jdc
from jaxtyping import Float, Array


@jdc.pytree_dataclass(kw_only=True)
class AbsorptionSpectrum:
    """Represents an absorption spectrum. Outputted by all `AbsorptionModel` subclasses.

    Parameters
    ----------
    energies : Float[Array, "num_points"]
        the x values of the absorption spectrum.
    intensities : Float[Array, "num_points"]
        the y values of the absorption spectrum.
    """

    energies: Float[Array, "num_points"]
    intensities: Float[Array, "num_points"]

    def save_data(self, filename: str) -> None:
        """Save the absorption spectrum data to a file.

        Parameters
        ----------
        filename : str
            output filename.
        """
        combined_data = np.column_stack(
            (np.array(self.energies), np.array(self.intensities))
        )

        np.savetxt(filename, combined_data, delimiter=",")

    def save_plot(self, filename: str) -> None:
        """Save the absorption spectrum plot to a file.

        Parameters
        ----------
        filename : str
            output filename.
        """
        plt.plot(self.energies, self.intensities)
        plt.xlabel("Energy (cm^-1)")
        plt.ylabel("Intensity")
        plt.savefig(filename)
        plt.close()



import jax.numpy as jnp
import jax_dataclasses as jdc

from quantumspectra_2024.common.absorption import (
    AbsorptionModel as Model,
    AbsorptionSpectrum,
)


@jdc.pytree_dataclass(kw_only=True)
class StarkModel(Model):
    """A general model for Stark absorption spectrum.

    Parameters
    ----------
    start_energy : float
        absorption spectrum's starting energy (wavenumbers).
    end_energy : float
        absorption spectrum's ending energy (wavenumbers).
    num_points : int
        absorption spectrum's number of points (unitless).

    neutral_submodel : Model
        parameterized neutral submodel to use in Stark effect calculation.

    positive_field_strength : float
        positive strength of the electric field.
    positive_field_sum_percent : float
        fraction of positive field strength to use in spectrum (decimal).
        negative field strength is 1 - this value.

    field_delta_dipole : float
        change in dipole moment due to electric field.
    field_delta_polarizability : float
        change in polarizability due to electric field.
    """

    #: parameterized neutral submodel to use in Stark effect calculation.
    neutral_submodel: Model

    #: positive strength of the electric field.
    positive_field_strength: float
    #: fraction of positive field strength to use in spectrum (decimal).
    positive_field_sum_percent: float = 0.5

    #: change in dipole moment due to electric field.
    field_delta_dipole: float
    #: change in polarizability due to electric field.
    field_delta_polarizability: float

    def get_absorption(self) -> AbsorptionSpectrum:
        """Compute the absorption spectrum for the model.

        First computes absorption spectrum for the neutral submodel, and then for the charged submodels.
        Two charged submodels are computed, one with positive field strength and one with negative field strength.

        A half-sum is computed between the two charged submodels, with each submodel's intensities scaled by
        `positive_field_sum_percent` and `1 - positive_field_sum_percent`, respectively.

        Then, the neutral intensities are subtracted from the charged half-sum to get the electroabsorption spectrum.

        Returns
        -------
        AbsorptionSpectrum
            the model's parameterized absorption spectrum.
        """
        # get absorption spectrum sample energies (x values)
        neutral_submodel = self.get_neutral_submodel()
        positive_submodel = self.get_charged_submodel(field_strength_scalar=1.0)
        negative_submodel = self.get_charged_submodel(field_strength_scalar=-1.0)

        neutral_absorption = neutral_submodel.get_absorption()

        neutral_spectrum = neutral_absorption.intensities
        positive_spectrum = (
            positive_submodel.get_absorption().intensities
            * self.positive_field_sum_percent
        )
        negative_spectrum = negative_submodel.get_absorption().intensities * (
            1 - self.positive_field_sum_percent
        )

        charged_half_sum = jnp.sum(
            jnp.array([positive_spectrum, negative_spectrum]), axis=0
        )

        electroabsorption_spectrum = charged_half_sum - jnp.array(neutral_spectrum)

        return AbsorptionSpectrum(
            energies=neutral_absorption.energies,
            intensities=electroabsorption_spectrum,
        )

    def get_neutral_submodel(self) -> Model:
        """Returns the netural submodel with the Stark effect applied.

        This method replaces the neutral submodel's point values with the Stark model's point values.
        No other changes are made to the neutral submodel.

        Returns
        -------
        Model
            the neutral submodel for Stark calculations.
        """
        # replace neutral submodel with own point values
        neutral_submodel = jdc.replace(
            self.neutral_submodel,
            start_energy=self.start_energy,
            end_energy=self.end_energy,
            num_points=self.num_points,
        )

        return neutral_submodel

    def get_charged_submodel(self, field_strength_scalar: float) -> Model:
        """Returns a charged submodel with the Stark effect applied.

        This method starts with the neutral submodel from `get_neutral_submodel` and applies the Stark effect.
        The Stark effect is applied through the Model's `apply_electric_field` method.
        Field strength is multiplied by `field_strength_scalar`, and all other values are inputted into the method exactly.

        Parameters
        ----------
        field_strength_scalar : float
            scalar to multiply the field strength by.

        Returns
        -------
        Model
            the charged submodel for Stark calculations.
        """
        # get neutral submodel
        neutral_submodel = self.get_neutral_submodel()

        # apply specified charge to neutral submodel
        field_strength = float(self.positive_field_strength * field_strength_scalar)
        field_delta_dipole = float(self.field_delta_dipole)
        field_delta_polarizability = float(self.field_delta_polarizability)

        charged_submodel = neutral_submodel.apply_electric_field(
            field_strength=field_strength,
            field_delta_dipole=field_delta_dipole,
            field_delta_polarizability=field_delta_polarizability,
        )

        return charged_submodel

    def apply_electric_field(*_) -> None:
        """Applies an electric field to the model. Stark effect is not configured to apply external electric fields.

        Raises
        ------
        NotImplementedError
            StarkModel does not support apply_electric_field.
        """
        raise NotImplementedError("StarkModel does not support apply_electric_field")



import math

import numpy as np
from jaxtyping import Float, Array


def calculate_mlj_spectrum(
    energy_gap: float,
    high_freq_frequency: float,
    high_freq_coupling: float,
    low_freq_frequency: float,
    low_freq_coupling: float,
    temperature_kelvin: float,
    disorder_meV: float,
    basis_size: int,
    sample_points: Float[Array, "num_points"],
) -> Float[Array, "num_points"]:
    """Computes an MLJ semiclassical absorption spectrum.

    Notes
    -----
    TODO:
    * Add more detailed implemention description.

    Parameters
    ----------
    energy_gap : float
        energy gap between states.
    high_freq_frequency : float
        frequency of high frequency mode.
    high_freq_coupling : float
        coupling of high frequency mode.
    low_freq_frequency : float
        frequency of low frequency mode.
    low_freq_coupling : float
        coupling of low frequency mode.
    temperature_kelvin : float
        temperature in Kelvin.
    disorder_meV : float
        disorder (sigma) in meV.
    basis_size : int
        size of basis set.
    sample_points : Float[Array, "num_points"]
        energy points to sample spectrum.

    Returns
    -------
    Float[Array, "num_points"]
        Computed absorption spectrum.
    """
    # compute necessary values
    disorder_wavenumbers = disorder_meV * 8061 * 0.001
    low_freq_relaxation_energy = low_freq_coupling**2 * low_freq_frequency
    temperature_kbT = temperature_kelvin * 0.695028
    high_freq_huang_rhys_factor = high_freq_coupling**2

    factorials = [math.factorial(n) for n in range(basis_size + 1)]

    def calculate_mlj_single_intensity(energy: float) -> float:
        abs_arr = [
            (
                np.exp(-high_freq_huang_rhys_factor)
                * (high_freq_huang_rhys_factor**n)
                / factorials[n]
            )
            * np.exp(
                -(
                    (
                        n * high_freq_frequency
                        + energy_gap
                        + low_freq_relaxation_energy
                        - energy
                    )
                    ** 2
                )
                / (
                    4 * low_freq_relaxation_energy * temperature_kbT
                    + 2 * disorder_wavenumbers**2
                )
            )
            for n in range(basis_size + 1)
        ]
        return np.sum(abs_arr)

    spectrum = np.vectorize(calculate_mlj_single_intensity)(sample_points)
    spectrum = spectrum / np.max(spectrum)
    return spectrum



import jax
import jax.numpy as jnp
from jaxtyping import Float, Array


def broaden_peaks(
    sample_points: Float[Array, "num_points"],
    peak_energies: Float[Array, "num_peaks"],
    peak_intensities: Float[Array, "num_peaks"],
    distribution_broadening: float,
) -> Float[Array, "num_points"]:
    """Broadens a set of peaks with given sample points into an absorption spectrum.

    This function first computes a set of Gaussian distributions centered at each peak energy, with a specified broadening.
    Then, each spectrum is scaled by the peak intensity and summed to form the final absorption spectrum.

    See `compute_gaussians` for more information on the Gaussian distributions.

    Parameters
    ----------
    sample_points : Float[Array, "num_points"]
        Sample points at which the absorption spectrum will be computed.
    peak_energies : Float[Array, "num_peaks"]
        Peak energies corresponding to the center of each Gaussian distribution.
    peak_intensities : Float[Array, "num_peaks"]
        Peak intensities to scale each Gaussian distribution.
    distribution_broadening : float
        Standard deviation of the Gaussian distributions.

    Returns
    -------
    Float[Array, "num_points"]
        Computed absorption spectrum.
    """
    gaussians = compute_gaussians(
        sample_points=sample_points,
        peak_energies=peak_energies,
        distribution_broadening=distribution_broadening,
    )
    return jnp.sum(peak_intensities * gaussians, axis=1)


@jax.jit
def compute_gaussians(
    sample_points: Float[Array, "num_points"],
    peak_energies: Float[Array, "num_peaks"],
    distribution_broadening: float,
) -> Float[Array, "num_points num_peaks"]:
    """Computes the Gaussian distributions for a set of sample points and peak energies.

    This function calculates the value of Gaussian functions for each combination of
    sample points and peak energies, considering a specified distribution broadening.
    It is `jax` JIT-compiled to improve performance.

    Parameters
    ----------
    sample_points : Float[Array, "num_points"]
        Sample points at which the Gaussian functions will be evaluated.
        Each point represents a position on the x-axis of the Gaussian distribution.
    peak_energies : Float[Array, "num_peaks"]
        Peak energies corresponding to the mean (center) of each Gaussian distribution.
        Each peak energy defines a different Gaussian function to be evaluated at the sample points.
    distribution_broadening : float
        A scalar value representing the standard deviation (sigma) of the Gaussian distributions,
        which determines the broadening of the distributions.
        This value is the same for all Gaussian distributions computed by this function.

    Returns
    -------
    Float[Array, "num_points num_peaks"]
        Computed Gaussian values for each combination of sample points and peak energies.
    """
    return (1.0 / (jnp.sqrt(jnp.pi) * distribution_broadening)) * jnp.exp(
        -jnp.power(
            (sample_points[:, jnp.newaxis] - peak_energies) / distribution_broadening, 2
        )
    )


def compute_peaks(
    eigenvalues: Float[Array, "matrix_size"],
    eigenvectors: Float[Array, "matrix_size matrix_size"],
    transfer_integral: float,
    temperature_kelvin: float,
) -> tuple[Float[Array, "num_peaks"], Float[Array, "num_peaks"]]:
    """Computes the absorption spectrum peak energies and intensities for a two-state system.

    Diagonalized eigenvalues and eigenvectors are first used to compute peak energies and intensities.
    See `compute_peak_energies` and `compute_peak_intensities` for more information.

    Intensities are scaled by probability scalars if the temperature is not 0.
    See `compute_peak_probability_scalars` for more information.

    Energies and intensities are then constrained only to pair combinations between a range of first eigenvectors and all other elevated energy levels.
    First eigenvector range is 1 with 0 temperature, or 50 with non-zero temperature.
    See `filter_peaks` for more information.

    Parameters
    ----------
    eigenvalues : Float[Array, "matrix_size"]
        Eigenvalues of the Hamiltonian.
    eigenvectors : Float[Array, "matrix_size matrix_size"]
        Eigenvectors of the Hamiltonian.
    transfer_integral : float
        Transfer integral between the two states.
    temperature_kelvin : float
        System's temperature in Kelvin.

    Returns
    -------
    tuple[Float[Array, "num_peaks"], Float[Array, "num_peaks"]]
        Computed absorption spectrum peak energies and intensities.
    """
    # compute all possible peak energies and intensities
    energies = compute_peak_energies(eigenvalues=eigenvalues)
    intensities = compute_peak_intensities(
        eigenvectors=eigenvectors, transfer_integral=transfer_integral
    )

    # compute temperature to wavenumbers
    temperature_wavenumbers = temperature_kelvin * 0.695028

    # define range of eigenvectors used in pair combinations
    first_eigenvector_range = jax.lax.cond(
        (temperature_wavenumbers == 0),
        lambda: 1,
        lambda: min(50, len(eigenvalues)),
    )

    # scale intensities by probability scalars if temperature is not 0
    scaled_intensities = jax.lax.cond(
        (temperature_wavenumbers == 0),
        lambda: intensities,
        lambda: intensities
        * compute_peak_probability_scalars(
            eigenvalues=eigenvalues, temperature_wavenumbers=temperature_wavenumbers
        )[:, None],
    )

    # filter computed intensities and energies to retrieve pair combinations between first_eigenvector_range and all other elevated energy levels
    filtered_energies, filtered_intensities = filter_peaks(
        peak_energies=energies,
        peak_intensities=scaled_intensities,
        first_eigenvector_range=first_eigenvector_range,
    )

    return filtered_energies, filtered_intensities


def compute_peak_energies(
    eigenvalues: Float[Array, "matrix_size"],
) -> Float[Array, "matrix_size/2 matrix_size"]:
    """Compute all raw spectrum peak energy values.

    Energy values are the differences between two state's eigenvalues.
    This function computes all differences between eigenvalues and returns a matrix of differences.
    The matrix follows the form where index i,j represents the difference between the ith and jth eigenvalues.

    Because only the first state will be used for pairs, only the first half of the differences are necessary and returned.

    Parameters
    ----------
    eigenvalues : Float[Array, "matrix_size"]
        Eigenvalues of the Hamiltonian.

    Returns
    -------
    Float[Array, "matrix_size/2 matrix_size"]
        difference matrix of eigenvalues.
    """
    # half the number of eigenvalues
    half_size = len(eigenvalues) // 2

    # reshape eigenvalues to column vector
    eigenvalues_col = eigenvalues[:, jnp.newaxis]
    # subtract the column vector from the transpose to get the difference matrix
    differences_matrix = eigenvalues - eigenvalues_col

    # only the first half is necessary
    return differences_matrix[:half_size]


def compute_peak_intensities(
    eigenvectors: Float[Array, "matrix_size matrix_size"],
    transfer_integral: float,
) -> Float[Array, "matrix_size/2 matrix_size"]:
    """Compute all raw spectrum peak intensity values.

    Intensity values are the dot product of two sets of sliced eigenvectors, squared.
    This function computes all squared dot products between pairs of eigenvectors and returns a matrix of intensities.
    The matrix follows the form where index i,j represents the intensity between the ith and jth eigenvectors.

    Because only the first state will be used for pairs, only the first half of the intensities are necessary and returned.

    The second eigenvalue in the computation is sliced in half depending on `transfer_integral` value:
        * if `transfer_integral` is not 0, the top half of the eigenvectors are used
        * if `transfer_integral` is 0, the bottom half of the eigenvectors are used

    Parameters
    ----------
    eigenvectors : Float[Array, "matrix_size matrix_size"]
        Eigenvectors of the Hamiltonian.
    transfer_integral : float
        Transfer integral between the two states.

    Returns
    -------
    Float[Array, "matrix_size/2 matrix_size"]
        intensity matrix of eigenvectors.
    """
    # half the size of a dimension of the eignvectors
    half_size = len(eigenvectors) // 2

    # slice the first dimension of the eigenvectors -> an array of half-sized vectors from the bottom of the eigenvectors
    vector_slices_1 = eigenvectors[:half_size, :]
    # slicing for the second set depends on t value -> if t is 0, slice the top half otherwise slice like vector_slices_1
    vector_slices_2 = jax.lax.cond(
        (transfer_integral == 0),
        lambda: eigenvectors[half_size:, :],
        lambda: vector_slices_1,
    )

    # compute dot product of the two sets of eigenvectors, and square result
    # first set is transposed to make result match dot products between all combinations of vectors
    intensities_matrix = jnp.dot(vector_slices_1.T, vector_slices_2) ** 2

    # only the first half is necessary
    return intensities_matrix[:half_size]


def compute_peak_probability_scalars(
    eigenvalues: Float[Array, "matrix_size"],
    temperature_wavenumbers: float,
) -> Float[Array, "matrix_size/2"]:
    """Computes a probability scalars for peak intensities.

    Computes a normalized exponential probability distribution based on the differences
    between the first half of the eigenvalues and the first eigenvalue, scaled by a given temperature.

    This function first calculates the differences between the first half of eigenvalues and the first eigenvalue.
    Then, it computes the exponential of the negative of these differences, divided by the temperature in wavenumbers.
    Finally, it normalizes these exponentials to sum to one, forming a probability distribution.

    The probability scalars are returned as a vector of the same size as the first half of the eigenvalues.
    The probability scalars are used to scale the intensities of the peaks in the absorption spectrum.

    Parameters
    ----------
    eigenvalues : Float[Array, "matrix_size"]
        Eigenvalues of the Hamiltonian.
    temperature_wavenumbers : float
        Temperature in wavenumbers.

    Returns
    -------
    Float[Array, "matrix_size/2"]
        probability scalars for peak intensities.
    """
    # half the number of eigenvalues
    half_size = len(eigenvalues) // 2

    # find differences between the first half of the eigenvalues and the first eigenvalue
    differences = eigenvalues[:half_size] - eigenvalues[0]
    # take exponentials of the negative differences divided by the temperature
    exponentials = jnp.exp(-differences / temperature_wavenumbers)

    # return normalized exponential of scaled differences
    return exponentials / jnp.sum(exponentials)


def filter_peaks(
    peak_energies: Float[Array, "matrix_size/2 matrix_size"],
    peak_intensities: Float[Array, "matrix_size/2 matrix_size"],
    first_eigenvector_range: int,
) -> tuple[Float[Array, "num_peaks"], Float[Array, "num_peaks"]]:
    """Filters peak energies and intensities by selecting unique and correct pair combinations and filtering negative values.

    This function uses upper-triangular matrix indices to select only peaks that have been computed with
    unique pair combinations between first eigenvectors in the given `first_eigenvector_range` and all other elevated energy levels.
    It also filters out negative energy or intensity values.

    Parameters
    ----------
    peak_energies : Float[Array, "matrix_size/2 matrix_size"]
        peak energies computed by `compute_peak_energies`.
    peak_intensities : Float[Array, "matrix_size/2 matrix_size"]
        peak intensities computed by `compute_peak_intensities`.
    first_eigenvector_range : int
        the range of first eigenvectors to consider for pair combinations.

    Returns
    -------
    tuple[Float[Array, "num_peaks"], Float[Array, "num_peaks"]]
        filtered peak energies and intensities.
    """
    # get upper-triangular indices starting from the first off-diagonal
    triu_indices = jnp.triu_indices(first_eigenvector_range, k=1, m=len(peak_energies))

    # define the filtering mask
    mask = ((peak_intensities >= 0) | (peak_energies >= 0))[triu_indices]

    # filter energies and intensities
    filtered_peak_energies = peak_energies[triu_indices][mask]
    filtered_peak_intensities = peak_intensities[triu_indices][mask]

    # this filters the arrays such that only unique pair combinations are considered, and
    # such that the intensities and energies are non-negative

    return filtered_peak_energies, filtered_peak_intensities



import jax
import jax.numpy as jnp
from jaxtyping import Float, Int, Bool, Array


def diagonalize_matrix(matrix: Float[Array, "matrix_size matrix_size"]) -> tuple[
    Float[Array, "matrix_size"],
    Float[Array, "matrix_size matrix_size"],
]:
    """Diagonalizes a matrix and returns the eigenvalues and eigenvectors in a tuple.

    Parameters
    ----------
    matrix : Float[Array, "matrix_size matrix_size"]
        a matrix to diagonalize.

    Returns
    -------
    tuple[Float[Array, "matrix_size"], Float[Array, "matrix_size matrix_size"]]
        tuple containing the eigenvalues and eigenvectors.
    """
    eigenvalues, eigenvectors = jnp.linalg.eigh(matrix)
    return eigenvalues, eigenvectors


def build_matrix(
    state_energies: Float[Array, "num_states"],
    transfer_integral: float,
    mode_basis_sets: Int[Array, "num_modes"],
    mode_localities: Bool[Array, "num_modes"],
    mode_frequencies: Float[Array, "num_modes"],
    mode_state_couplings: Float[Array, "num_modes num_states"],
) -> Float[Array, "matrix_size matrix_size"]:
    """Builds a full Hamiltonian matrix for a system of states and modes.

    Hamiltonians are built with a combination of blocks. Each block represents a single state.
    States along the diagonal are so-called 'local' states, while offdiagonal states are 'nonlocal'.
    See documentation for `build_local_state_block` and `build_nonlocal_state_block` for more information.

    Notes
    -----
    TODO:
    * Implement multiple transfer integrals for more than two states.

    Parameters
    ----------
    state_energies : Float[Array, "num_states"]
        energies of all local states.
    transfer_integral : float
        transfer integral between states.
    mode_basis_sets : Int[Array, "num_modes"]
        basis set size per mode.
    mode_localities : Bool[Array, "num_modes"]
        locality of each mode.
    mode_frequencies : Float[Array, "num_modes"]
        frequency per mode.
    mode_state_couplings : Float[Array, "num_modes num_states"]
        coupling per mode and local state.

    Returns
    -------
    Float[Array, "matrix_size matrix_size"]
        fully constructed matrix in `jax` array.
    """
    num_states = len(state_energies)

    # build the matrix, state by state
    rows = []
    for state_row in range(num_states):
        cols = []
        for state_col in range(num_states):
            state_index = max(state_row, state_col)

            if state_row == state_col:
                # calculate a local state block
                state = build_local_state_block(
                    state_index=state_index,
                    state_energies=state_energies,
                    mode_basis_sets=mode_basis_sets,
                    mode_localities=mode_localities,
                    mode_frequencies=mode_frequencies,
                    mode_state_couplings=mode_state_couplings,
                )
            else:
                # calculate a nonlocal state block
                state = build_nonlocal_state_block(
                    state_index=state_index,
                    transfer_integral=transfer_integral,
                    mode_basis_sets=mode_basis_sets,
                    mode_localities=mode_localities,
                    mode_frequencies=mode_frequencies,
                    mode_state_couplings=mode_state_couplings,
                )
            cols.append(state)
        rows.append(jnp.hstack(cols))
    matrix = jnp.vstack(rows)

    return matrix


def build_local_state_block(
    state_index: int,
    state_energies: Float[Array, "num_states"],
    mode_basis_sets: Int[Array, "num_modes"],
    mode_localities: Bool[Array, "num_modes"],
    mode_frequencies: Float[Array, "num_modes"],
    mode_state_couplings: Float[Array, "num_modes num_states"],
) -> Float[Array, "block_size block_size"]:
    """Builds a local state block.

    To see how block are built, see `build_state_block`.
    To build a local state block, the following steps are taken.
    * Calculate the state's diagonal values with `calculate_state_local_diagonals`.
    * Calculate the state's offdiagonal values with `calculate_state_offdiagonals`.
    * Build the state block.

    Parameters
    ----------
    state_index : int
        index of the state.
    state_energies : Float[Array, "num_states"]
        energies of all local states.
    mode_basis_sets : Int[Array, "num_modes"]
        basis set size per mode.
    mode_localities : Bool[Array, "num_modes"]
        locality of each mode.
    mode_frequencies : Float[Array, "num_modes"]
        frequency per mode.
    mode_state_couplings : Float[Array, "num_modes num_states"]
        coupling per mode and local state.

    Returns
    -------
    Float[Array, "block_size block_size"]
        constructed matrix block.
    """
    state_energy = state_energies[state_index]
    mode_couplings = mode_state_couplings[:, state_index]

    # calculate the state's diagonal values
    all_diagonal_values = calculate_state_local_diagonals(
        state_energy=state_energy,
        mode_frequencies=mode_frequencies,
        mode_couplings=mode_couplings,
        mode_basis_sets=mode_basis_sets,
    )

    # calculate the state's offdiagonal values, arranged in a tuple for each mode
    all_mode_offdiagonal_values = calculate_state_offdiagonals(
        state_locality=True,
        mode_basis_sets=mode_basis_sets,
        mode_localities=mode_localities,
        mode_frequencies=mode_frequencies,
        mode_couplings=mode_couplings,
    )

    # build a state block with the diagonal values and tuple of offdiagonal values
    return build_state_block(
        all_diagonal_values=all_diagonal_values,
        all_mode_offdiagonal_values=all_mode_offdiagonal_values,
        mode_basis_sets=mode_basis_sets,
    )


def build_nonlocal_state_block(
    state_index: int,
    transfer_integral: float,
    mode_basis_sets: Int[Array, "num_modes"],
    mode_localities: Bool[Array, "num_modes"],
    mode_frequencies: Float[Array, "num_modes"],
    mode_state_couplings: Float[Array, "num_modes num_states"],
):
    """Builds a nonlocal state block.

    To see how blocks are built, see `build_state_block`.
    To build a nonlocal state block, the following steps are taken.
    * The state's offdiagonal values are all set to the transfer integral.
    * Calculate the state's diagonal values with `calculate_state_local_diagonals`.
    * Build the state block.

    Parameters
    ----------
    state_index : int
        index of the state.
    transfer_integral : float
        transfer integral between states.
    mode_basis_sets : Int[Array, "num_modes"]
        basis set size per mode.
    mode_localities : Bool[Array, "num_modes"]
        locality of each mode.
    mode_frequencies : Float[Array, "num_modes"]
        frequency per mode.
    mode_state_couplings : Float[Array, "num_modes num_states"]
        coupling per mode and local state.

    Returns
    -------
    Float[Array, "block_size block_size"]
        constructed matrix block.
    """
    mode_couplings = mode_state_couplings[:, state_index]

    # calculate the state's diagonal values
    all_diagonal_values = jnp.repeat(
        transfer_integral, jnp.prod(jnp.array(mode_basis_sets))
    )

    # calculate the state's offdiagnoal values, arranged in a tuple for each mode
    all_mode_offdiagonal_values = calculate_state_offdiagonals(
        state_locality=False,
        mode_basis_sets=mode_basis_sets,
        mode_localities=mode_localities,
        mode_frequencies=mode_frequencies,
        mode_couplings=mode_couplings,
    )

    # build a state block with the diagonal values and tuple of offdiagonal values
    return build_state_block(
        all_diagonal_values=all_diagonal_values,
        all_mode_offdiagonal_values=all_mode_offdiagonal_values,
        mode_basis_sets=mode_basis_sets,
    )


def build_state_block(
    all_diagonal_values: Float[Array, "block_size"],
    all_mode_offdiagonal_values: tuple[Float[Array, "_"]],
    mode_basis_sets: Float[Array, "num_modes"],
) -> Float[Array, "block_size block_size"]:
    """Builds a single block of the full Hamiltonian matrix for a single state.

    Blocks have the form:
        - diagonal values go across the main diagonal
        - state blocks are broken into inner blocks for each state.
            From there, the diagonals one above and below the main diagonal (designated 'offdiagonals')
            are filled with values corresponding to the mode's offdiagonal components, where the diagonal's
            lower index is used in generation. Offdiagonal values are repeated to evenly fill the offdiagonal.

            For instance, two modes of basis set 3 will create the following block:

            [ d_0, m_0, 0,   n_0, 0,   0,   0,   0,   0
              m_0, d_1, m_1, 0,   n_0, 0,   0,   0,   0
              0,   m_1, d_2, 0,   0,   n_0, 0,   0,   0
              n_0, 0,   0,   d_3, m_0, 0,   n_1, 0,   0
              0,   n_0, 0,   m_0, d_4, m_1, 0,   n_1, 0
              0,   0,   n_0, 0,   m_1, d_5, 0,   0,   n_1
              0,   0,   0,   n_1, 0,   0, d_6,   m_0, 0
              0,   0,   0,   0,   n_1, 0,   m_0, d_7, m_1
              0,   0,   0,   0,   0,   n_1, 0,   m_1, d_8 ]

              Where d_i is the diagonal value at index i,
              m_i is the offdiagonal value for the last mode at subindex i
              n_i is the offdiagonal value for the last mode at subindex i

    Parameters
    ----------
    all_diagonal_values : Float[Array, "block_size"]
        all state diagonal values.
    all_mode_offdiagonal_values : tuple[Float[Array, "_"]]
        all offdiagonal values for the state block per mode.
    mode_basis_sets : Float[Array, "num_modes"]
        basis set sizes per mode.

    Returns
    -------
    Float[Array, "block_size block_size"]
        constructed matrix block.
    """
    # start with an empty block of size 1
    block = jnp.zeros((1, 1))

    # run recursively for each mode, expanding by the basis set size and filling offdiagonal values
    for mode_basis_set, mode_offdiagonal_values in zip(
        reversed(mode_basis_sets), reversed(all_mode_offdiagonal_values)
    ):
        mode_offdiagonal_values = jnp.array(mode_offdiagonal_values)
        previous_block_size = len(block)

        # repeat each value to match the previous block size
        mode_offdiagonal_values = jnp.repeat(
            mode_offdiagonal_values, repeats=previous_block_size
        )

        # create a new block by repeating the previous block to match current basis set size
        new_block = jax.scipy.linalg.block_diag(*[block for _ in range(mode_basis_set)])

        # redefines the block by combining the new block with the new offdiagonal values
        block = (
            new_block
            + jnp.diag(mode_offdiagonal_values, k=previous_block_size)
            + jnp.diag(mode_offdiagonal_values, k=-previous_block_size)
        )

    # finally, fills the main diagonal of the full block
    block = block + jnp.diag(all_diagonal_values)
    return block


def calculate_state_local_diagonals(
    state_energy: float,
    mode_frequencies: Float[Array, "num_modes"],
    mode_couplings: Float[Array, "num_modes"],
    mode_basis_sets: Int[Array, "num_modes"],
) -> Float[Array, "block_size"]:
    """Calculate all diagonal values for a state block in a single array. Only calculated for local blocks.

    Diagonal values are calculated with a sum of contributions from each mode, plus the state energy.
    Contributions for each mode are summed with all combinations of values between 1 and their basis set size - 1.
    To see how individual mode contributions are calculated, see `calculate_mode_local_diagonal_component`.

    Parameters
    ----------
    state_energy : float
        energy of the state.
    mode_frequencies : Float[Array, "num_modes"]
        frequencies of each mode.
    mode_couplings : Float[Array, "num_modes"]
        couplings of each mode.
    mode_basis_sets : Int[Array, "num_modes"]
        basis set size of each mode.

    Returns
    -------
    Float[Array, "block_size"]
        all diagonal values for the state block.
    """
    parallelized_contribution_func = jax.vmap(
        calculate_mode_local_diagonal_component, in_axes=(0, None, None)
    )

    mode_diagonal_contributions = [
        parallelized_contribution_func(
            jnp.arange(mode_basis_set),
            mode_frequency,
            mode_coupling,
        )
        for mode_basis_set, mode_frequency, mode_coupling in zip(
            mode_basis_sets, mode_frequencies, mode_couplings
        )
    ]

    sum_contribution_combinations = outer_sum(*mode_diagonal_contributions).flatten()

    all_diagonal_values = state_energy + sum_contribution_combinations

    return all_diagonal_values


def calculate_state_offdiagonals(
    state_locality: bool,
    mode_basis_sets: Int[Array, "num_modes"],
    mode_localities: Bool[Array, "num_modes"],
    mode_frequencies: Float[Array, "num_modes"],
    mode_couplings: Float[Array, "num_modes"],
) -> tuple[Float[Array, "_mode_offdiagonal_size"]]:
    """Calculate all unqiue offdiagonal values for a state block per mode. Each mode's set of offdiagonals is contained in a tuple.

    Offdiagonal values are an array of calculated values that range from 1 to the basis set size of the mode - 1.
    Thus, each array in the tuple will have a length of `mode_basis_set`.
    If the mode doesn't match the locality of the state, the offdiagonals are all set to zero.
    To see how individual mode offdiagonal values are calculated, see `calculate_mode_offdiagonal_component`.

    Parameters
    ----------
    state_locality : bool
        the locality of the state.
    mode_basis_sets : Int[Array, "num_modes"]
        basis set size of each mode.
    mode_localities : Bool[Array, "num_modes"]
        locality of each mode.
    mode_frequencies : Float[Array, "num_modes"]
        frequencies of each mode.
    mode_couplings : Float[Array, "num_modes"]
        couplings of each mode.

    Returns
    -------
    tuple[Float[Array, "_mode_offdiagonal_size"]]
        all offdiagonal values for the state block per mode.
    """
    all_mode_offdiagonal_values = tuple(
        (
            [
                calculate_mode_offdiagonal_component(
                    component_index=component_index,
                    mode_frequency=mode_frequency,
                    mode_coupling=mode_coupling,
                )
                for component_index in range(mode_basis_set - 1)
            ]
            if mode_locality == state_locality
            else jnp.zeros((mode_basis_set - 1,))
        )
        for mode_locality, mode_basis_set, mode_frequency, mode_coupling in zip(
            mode_localities, mode_basis_sets, mode_frequencies, mode_couplings
        )
    )

    return all_mode_offdiagonal_values


def calculate_mode_local_diagonal_component(
    component_index: int,
    mode_frequency: float,
    mode_coupling: float,
) -> float:
    """Computes a single diagonal contribution component for a mode.

    Diagonal contributions are calculated as follows:
        * The mode frequency is multiplied by the component index plus one half
        * This value is then multiplied by the square of the mode coupling, divided by two.

    Parameters
    ----------
    component_index : int
        the index of the component.
    mode_frequency : float
        the frequency of the mode.
    mode_coupling : float
        the coupling of the mode.

    Returns
    -------
    float
        the computed diagonal contribution component.
    """
    return mode_frequency * ((component_index + 0.5) + (mode_coupling**2))


def calculate_mode_offdiagonal_component(
    component_index: int,
    mode_frequency: float,
    mode_coupling: float,
) -> float:
    """Computes a single offdiagonal contribution component for a mode.

    Offdiagonal contributions are calculated as follows:
        * The mode frequency is multiplied by the square root of the component index plus one, divided by two.
        * This value is then multiplied by the mode coupling.

    Parameters
    ----------
    component_index : int
        the index of the component.
    mode_frequency : float
        the frequency of the mode.
    mode_coupling : float
        the coupling of the mode.

    Returns
    -------
    float
        the computed offdiagonal contribution component.
    """
    return mode_frequency * mode_coupling * jnp.sqrt(component_index + 1)


def outer_sum(*arrays: tuple[Float[Array, "*"]]) -> Float[Array, "*"]:
    """
    Computes the outer sum of multiple JAX arrays.

    This function takes multiple JAX arrays as input and computes their outer sum. It starts with the first array and
    iteratively adds each subsequent array to it in a way that's similar to computing the outer product, but with
    summation instead. This is done by reshaping the arrays for broadcasting, ensuring dimensions are aligned correctly
    for the sum.

    Parameters
    ----------
    *arrays: tuple[Float[Array, "*"]]
        Variable number of JAX array arguments.
        Each array should be compatible for broadcasting.
        There should be at least one array passed to this function.

    Returns
    -------
    Float[Array, "*"]
        A JAX array containing the outer sum of the input arrays.

    Raises
    ------
    ValueError
        If no arrays are provided as input.

    Examples
    --------
        >>> import jax.numpy as jnp
        >>> a = jnp.array([1, 2])
        >>> b = jnp.array([3, 4])
        >>> outer_sum(a, b)
        DeviceArray([[4, 5],
                     [5, 6]], dtype=int32)

    Notes
    -----
        The function requires at least one input array and all input arrays must be compatible for broadcasting
        following the JAX rules.
    """
    # Ensure there is at least one array
    if not arrays:
        raise ValueError("At least one array is required")

    # Start with the first array, reshaping it to have a new axis for each additional array
    result = arrays[0].reshape(arrays[0].shape + (1,) * (len(arrays) - 1))

    # Iteratively add each subsequent array, reshaping appropriately for broadcasting
    for i, arr in enumerate(arrays[1:], 1):
        # The new shape has 1's in all positions except the current dimension being added
        new_shape = (1,) * i + arr.shape + (1,) * (len(arrays) - i - 1)
        result += arr.reshape(new_shape)

    return result



import jax.numpy as jnp
import jax_dataclasses as jdc
from jaxtyping import Float, Int, Array

from quantumspectra_2024.common.absorption import (
    AbsorptionModel as Model,
    AbsorptionSpectrum,
)
from quantumspectra_2024.common.hamiltonian import HamiltonianModel
from quantumspectra_2024.models.two_state.TwoStateComputation import (
    compute_peaks,
    broaden_peaks,
)


@jdc.pytree_dataclass(kw_only=True)
class TwoStateModel(Model):
    """A two-state quantum mechanical model for absorption spectra.

    Parameters
    ----------
    start_energy : float
        absorption spectrum's starting energy (wavenumbers).
    end_energy : float
        absorption spectrum's ending energy (wavenumbers).
    num_points : int
        absorption spectrum's number of points (unitless).

    broadening: float
        absorption spectrum broadening factor (wavenumbers).
    temperature_kelvin : float
        system's temperature (Kelvin).

    transfer_integral : float
        transfer integral between the two states.
    energy_gap : float
        energy gap between the two states (wavenumbers).

    mode_basis_sets : Float[Array, "num_modes"]
        basis set size per mode (unitless).
    mode_frequencies : Float[Array, "num_modes"]
        frequency per mode (wavenumbers).
    mode_couplings : Float[Array, "num_modes"]
        excited state coupling per mode.
    """

    #: absorption spectrum broadening factor (wavenumbers).
    broadening: float = 200.0
    #: system's temperature (Kelvin).
    temperature_kelvin: float

    #: transfer integral between the two states.
    transfer_integral: float
    #: energy gap between the two states (wavenumbers).
    energy_gap: float

    #: basis set size per mode (unitless).
    mode_basis_sets: Int[Array, "num_modes"]
    #: frequency per mode (wavenumbers).
    mode_frequencies: Float[Array, "num_modes"]
    #: excited state coupling per mode.
    mode_couplings: Float[Array, "num_modes"]

    def get_absorption(self) -> AbsorptionSpectrum:
        """Compute the absorption spectrum for the model.

        First computes the Hamiltonian, then diagonalizes it to get eigenvalues and eigenvectors.
        Then computes the absorption spectrum peaks and broadens them into a spectrum.

        See docs in :class:`quantumspectra_2024.modules.hamiltonian.HamiltonianComputation` and :class:`TwoStateComputation` to see how this is done.

        Returns
        -------
        AbsorptionSpectrum
            the model's parameterized absorption spectrum.
        """
        # compute the Hamiltonian
        hamiltonian = self.get_hamiltonian()

        # diagonalize the Hamiltonian
        eigenvalues, eigenvectors = hamiltonian.get_diagonalization()

        # get absorption spectrum sample energies (x values)
        sample_points = jnp.linspace(
            self.start_energy, self.end_energy, self.num_points
        )

        # compute absorption spectrum peaks
        peak_energies, peak_intensities = compute_peaks(
            eigenvalues=eigenvalues,
            eigenvectors=eigenvectors,
            transfer_integral=self.transfer_integral,
            temperature_kelvin=self.temperature_kelvin,
        )

        # broaden peaks into spectrum
        spectrum = broaden_peaks(
            sample_points=sample_points,
            peak_energies=peak_energies,
            peak_intensities=peak_intensities,
            distribution_broadening=self.broadening,
        )

        # return as AbsorptionSpectrum dataclass
        return AbsorptionSpectrum(
            energies=sample_points,
            intensities=spectrum,
        )

    def get_hamiltonian(self) -> HamiltonianModel:
        """Returns the model's associated `HamiltonianModel`.

        Returns
        -------
        HamiltonianModel
            the model's Hamiltonian.
        """
        return HamiltonianModel(
            transfer_integral=self.transfer_integral,
            state_energies=jnp.array([0.0, self.energy_gap]),
            mode_basis_sets=jnp.array(self.mode_basis_sets),
            mode_localities=jnp.array([True, True]),
            mode_frequencies=jnp.array(self.mode_frequencies),
            mode_state_couplings=jnp.array(
                [[0.0, mode_coupling] for mode_coupling in self.mode_couplings]
            ),
        )

    def apply_electric_field(
        self,
        field_strength: float,
        field_delta_dipole: float,
        field_delta_polarizability: float,
    ) -> "TwoStateModel":
        """Applies an electric field to the model. Returns a new instance of the model.

        Parameters
        ----------
        field_strength : float
            the strength of the electric field.
        field_delta_dipole : float
            the change in dipole moment due to the electric field.
        field_delta_polarizability : float
            the change in polarizability due to the electric field.

        Returns
        -------
        TwoStateModel
            the model with the electric field applied.
        """
        dipole_energy_change = field_delta_dipole * field_strength * 1679.0870295
        polarizability_energy_change = (
            0.5 * (field_strength**2) * field_delta_polarizability * 559.91
        )
        field_energy_change = -1 * (dipole_energy_change + polarizability_energy_change)

        return jdc.replace(
            self,
            energy_gap=self.energy_gap + field_energy_change,
        )



from quantumspectra_2024.common.Config import (
    parse_config,
    initialize_absorption_from_config,
    save_spectrum_from_config,
)
from quantumspectra_2024.common.absorption import (
    AbsorptionModel,
    AbsorptionSpectrum,
)

from quantumspectra_2024.models import TwoStateModel, MLJModel, StarkModel


def main():
    config: dict = parse_config("Compute absorption spectrum with a given config file.")

    str_to_model: dict = {
        "two_state": TwoStateModel,
        "mlj": MLJModel,
        "stark": StarkModel,
    }
    model: AbsorptionModel = initialize_absorption_from_config(
        config=config, str_to_model=str_to_model
    )

    spectrum: AbsorptionSpectrum = model.get_absorption()

    save_spectrum_from_config(config=config, spectrum=spectrum)


if __name__ == "__main__":
    main()



import jax_dataclasses as jdc
from jaxtyping import Float, Int, Bool, Array

from quantumspectra_2024.common.hamiltonian.HamiltonianComputation import (
    build_matrix,
    diagonalize_matrix,
)


@jdc.pytree_dataclass(kw_only=True)
class HamiltonianModel:
    """A hamiltonian model for a quantum system.

    Attributes
    ----------
    transfer_integral : float
        transfer integral between states.
    state_energies : Float[Array, "num_states"]
        energies of each state.

    mode_basis_sets : Int[Array, "num_modes"]
        basis set size per mode.
    mode_localities : Bool[Array, "num_modes"]
        whether each mode is local.
    mode_frequencies : Float[Array, "num_modes"]
        frequency per mode.
    mode_state_couplings : Float[Array, "num_modes num_states"]
        coupling per mode and state.
    """

    #: transfer integral between states.
    transfer_integral: float
    #: energies of each state.
    state_energies: Float[Array, "num_states"]

    #: basis set size per mode.
    mode_basis_sets: Int[Array, "num_modes"]
    #: whether each mode is local.
    mode_localities: Bool[Array, "num_modes"]
    #: frequency per mode.
    mode_frequencies: Float[Array, "num_modes"]
    #: coupling per mode and state.
    mode_state_couplings: Float[Array, "num_modes num_states"]

    def get_diagonalization(self) -> tuple[
        Float[Array, "matrix_size"],
        Float[Array, "matrix_size matrix_size"],
    ]:
        """Compute the diagonalization of the Hamiltonian.

        First computes the Hamiltonian matrix, then diagonalizes it.
        See docs in `HamiltonianComputation` to see how this is done.

        Returns
        -------
        tuple[Float[Array, "matrix_size"], Float[Array, "matrix_size matrix_size"]]
            tuple containing the eigenvalues and eigenvectors in `jax` arrays.
        """
        # build matrix
        matrix = self.get_matrix()

        # diagonalize matrix
        eigenvalues, eigenvectors = diagonalize_matrix(matrix)

        return eigenvalues, eigenvectors

    def get_matrix(self) -> Float[Array, "matrix_size matrix_size"]:
        """Builds the Hamiltonian matrix.

        See docs in `HamiltonianComputation` to see how this is done.

        Returns
        -------
        Float[Array, "matrix_size matrix_size"]
            fully constructed matrix in `jax` array.
        """
        return build_matrix(
            mode_basis_sets=self.mode_basis_sets,
            state_energies=self.state_energies,
            transfer_integral=self.transfer_integral,
            mode_localities=self.mode_localities,
            mode_frequencies=self.mode_frequencies,
            mode_state_couplings=self.mode_state_couplings,
        )



from typing import Type, Callable
from pathlib import Path
from argparse import ArgumentParser
import tomllib

from quantumspectra_2024.common.absorption import (
    AbsorptionModel,
    AbsorptionSpectrum,
)

CONFIG_ARG_NAME = "config_path"
CONFIG_ARG_HELP = "Path to the configuration file."

OUT_CONFIG_NAME = "out"
OUT_REQUIRED_KEYS = ["filename", "data", "plot", "overwrite"]

MODEL_CONFIG_NAME = "model"
MODEL_NAME_KEY = "name"


def save_spectrum_from_config(config: dict, spectrum: AbsorptionSpectrum) -> None:
    """Saves an absorption spectrum to requested files based on a configuration dict.

    Currently configured to save absorption spectrum data and plot to CSV and PNG files, respectively.

    Parameters
    ----------
    config : dict
        the configuration dict.
    spectrum : AbsorptionSpectrum
        the absorption spectrum to save.
    """
    out_data = config[OUT_CONFIG_NAME]
    overwrite = out_data["overwrite"]

    if out_data["data"]:
        # save absorption spectrum data
        save_file(
            filename=f"{out_data['filename']}.csv",
            overwrite=overwrite,
            save_func=lambda fname: spectrum.save_data(fname),
        )

    if out_data["plot"]:
        # save absorption spectrum plot
        save_file(
            filename=f"{out_data['filename']}.png",
            overwrite=overwrite,
            save_func=lambda fname: spectrum.save_plot(fname),
        )


def save_file(filename: str, overwrite: bool, save_func: Callable[[str], None]) -> None:
    """Checks if a file can be saved and then saves it using a save function.

    Parameters
    ----------
    filename : str
        the filename to save to.
    overwrite : bool
        whether to overwrite the file if it already exists.
    save_func : Callable[[str], None]
        The function to use to save the file.
        Function expects a single string argument containing the path to save to.

    Raises
    ------
    ValueError
        if the parent directory of the save file does not exist.
    ValueError
        if the save file already exists and overwrite is set to False.
    """
    file = Path(filename)

    if not file.parent.exists():
        raise ValueError(
            f"Invalid save file: parent directory '{file.parent}' does not exist."
        )
    if file.exists() and not overwrite:
        raise ValueError(
            f"Save file '{file}' already exists and overwrite is set to False."
        )

    save_func(str(file))


def initialize_absorption_from_config(
    config: dict, str_to_model: dict[str, Type[AbsorptionModel]]
) -> AbsorptionModel:
    """Initializes an `AbsorptionModel` subclass from a configuration dict.

    Configuration dicts must have a `model` key with a `name` key specifying the model name.
    The model name must be a key in the `str_to_model` dict.

    Model subdicts are treated as submodels and are recursively initialized.
    This is useful for models that have submodels as attributes, such as `StarkModel`.

    Parameters
    ----------
    config : dict
        the configuration dict.
    str_to_model dict[str, Type[AbsorptionModel]]
        a dict mapping model names to model classes.

    Raises
    ------
    ValueError
        if the model name is not found in `str_to_model`.

    Returns
    -------
    AbsorptionModel
        the initialized absorption model.
    """
    model_config = config[MODEL_CONFIG_NAME]
    model_name = model_config[MODEL_NAME_KEY]

    if model_name not in str_to_model:
        raise ValueError(
            f"Config model name '{model_name}' not recognized. "
            f"Available models: {','.join(str_to_model.keys())}"
        )

    model = str_to_model[model_name]

    model_config.pop(MODEL_NAME_KEY)

    # replace subdicts with submodels
    for key, value in model_config.items():
        if isinstance(value, dict):
            submodel = initialize_absorption_from_config(
                config={MODEL_CONFIG_NAME: value, OUT_CONFIG_NAME: {}},
                str_to_model=str_to_model,
            )
            model_config[key] = submodel

    return model(**model_config)


def parse_config(program_description: str) -> dict:
    """Parse and validate a configuration dict for absorption spectrum computation from script arguments.

    Parameters
    ----------
    program_description : str
        the program description for the argument parser.

    Returns
    -------
    dict
        the absorption spectrum computation configuration dict."""
    parser = ArgumentParser(description=program_description)

    parser.add_argument(CONFIG_ARG_NAME, type=str, help=CONFIG_ARG_HELP)
    config_path = parser.parse_args().config_path

    with open(config_path, "rb") as f:
        data = tomllib.load(f)

    validate_config(data)

    return data


def validate_config(data: dict) -> None:
    """Validates an absorption spectrum computation configuration dict.

    Parameters
    ----------
    data : dict
        the configuration dict to validate.

    Raises
    ------
    ValueError
        if the config is not valid.
    """
    # check for required keys
    required_keys = [MODEL_CONFIG_NAME, OUT_CONFIG_NAME]
    ensure_keys_included(data, required_keys, "main")

    # check for required output keys
    out_data = data[OUT_CONFIG_NAME]
    ensure_keys_included(out_data, OUT_REQUIRED_KEYS, OUT_CONFIG_NAME)

    # check for required model keys
    model_data = data[MODEL_CONFIG_NAME]
    model_required_keys = [MODEL_NAME_KEY]
    ensure_keys_included(model_data, model_required_keys, MODEL_CONFIG_NAME)


def ensure_keys_included(data: dict, keys: list, key_type: str) -> None:
    """Ensures that a list of required keys are included in a data dict.

    Parameters
    ----------
    data : dict
        the data dict to check.
    keys : list
        the list of required keys.
    key_type : str
        the type of data dict being checked.

    Raises
    ------
    ValueError
        if a required key is not found in the data dict.
    """
    for key in keys:
        if key not in data:
            raise ValueError(
                f"Required config key '{key}' not found in {key_type} data."
            )



import numpy as np
import jax.numpy as jnp
import jax_dataclasses as jdc
from jaxtyping import Float, Array

from quantumspectra_2024.common.absorption import (
    AbsorptionModel as Model,
    AbsorptionSpectrum,
)
from quantumspectra_2024.models.mlj.MLJComputation import calculate_mlj_spectrum


@jdc.pytree_dataclass(kw_only=True)
class MLJModel(Model):
    """A two-state two-mode MLJ model for absorption spectra.

    Parameters
    ----------
    start_energy : float
        absorption spectrum's starting energy (wavenumbers).
    end_energy : float
        absorption spectrum's ending energy (wavenumbers).
    num_points : int
        absorption spectrum's number of points (unitless).

    temperature_kelvin : float
        system's temperature (Kelvin).
    energy_gap : float
        energy gap between the two states (wavenumbers).
    disorder_meV : float
        disorder in the system (meV).

    basis_size : int
        size of basis set (unitless).

    mode_frequencies : Float[Array, "2"]
        frequency per mode (wavenumbers).
    mode_couplings : Float[Array, "2"]
        excited state coupling per mode.
    """

    #: system's temperature (Kelvin).
    temperature_kelvin: float
    #: energy gap between the two states (wavenumbers).
    energy_gap: float
    #: disorder in the system (meV).
    disorder_meV: float

    #: size of basis set (unitless).
    basis_size: int = 20

    #: frequency per mode (wavenumbers), must have exactly two.
    mode_frequencies: Float[Array, "2"]
    #: excited state coupling per mode, must have exactly two.
    mode_couplings: Float[Array, "2"]

    def get_absorption(self) -> AbsorptionSpectrum:
        """Compute the absorption spectrum for the model.

        See docs in `MLJComputation` to see how this is done.

        Returns
        -------
        AbsorptionSpectrum
            the model's parameterized absorption spectrum.
        """
        # get absorption spectrum sample energies (x values)
        sample_points = jnp.linspace(
            float(self.start_energy), float(self.end_energy), int(self.num_points)
        )

        # get low and high frequency modes
        lower_frequency, lower_coupling, higher_frequency, higher_coupling = (
            self.get_low_high_frequency_modes()
        )

        # calculate absorption spectrum
        spectrum = calculate_mlj_spectrum(
            energy_gap=float(self.energy_gap),
            high_freq_frequency=float(higher_frequency),
            high_freq_coupling=float(higher_coupling),
            low_freq_frequency=float(lower_frequency),
            low_freq_coupling=float(lower_coupling),
            temperature_kelvin=float(self.temperature_kelvin),
            disorder_meV=float(self.disorder_meV),
            basis_size=int(self.basis_size),
            sample_points=np.array(sample_points),
        )

        # return as AbsorptionSpectrum dataclass
        return AbsorptionSpectrum(
            energies=sample_points,
            intensities=jnp.array(spectrum),
        )

    def get_low_high_frequency_modes(self) -> tuple[float, float, float, float]:
        """Extracts the low and high frequency modes and mode couplings from the model.

        Model was already verified to have exactly two modes.
        This function sorts the modes by frequency and returns the lowest and highest frequency modes.
        This is useful for the `calculate_mlj_spectrum` function, which expects the low and high frequency modes explicitly.

        Returns
        -------
        tuple[float, float, float, float]
            low frequency, low coupling, high frequency, high coupling.
        """
        mode_frequencies = np.array(self.mode_frequencies)
        mode_couplings = np.array(self.mode_couplings)

        sorted_frequency_indices = np.argsort(mode_frequencies)

        sorted_frequencies = mode_frequencies[sorted_frequency_indices]
        sorted_couplings = mode_couplings[sorted_frequency_indices]

        # Assign to variables
        lower_frequency = sorted_frequencies[0]
        lower_coupling = sorted_couplings[0]
        higher_frequency = sorted_frequencies[-1]
        higher_coupling = sorted_couplings[-1]

        return lower_frequency, lower_coupling, higher_frequency, higher_coupling

    def apply_electric_field(
        self,
        field_strength: float,
        field_delta_dipole: float,
        field_delta_polarizability: float,
    ) -> "MLJModel":
        """Applies an electric field to the model. Returns a new instance of the model.

        Parameters
        ----------
        field_strength : float
            the strength of the electric field.
        field_delta_dipole : float
            the change in dipole moment due to the electric field.
        field_delta_polarizability : float
            the change in polarizability due to the electric field.

        Returns
        -------
        MLJModel
            the model with the electric field applied.
        """
        dipole_energy_change = field_delta_dipole * field_strength * 1679.0870295
        polarizability_energy_change = (
            0.5 * (field_strength**2) * field_delta_polarizability * 559.91
        )
        field_energy_change = -1 * (dipole_energy_change + polarizability_energy_change)

        return jdc.replace(
            self,
            energy_gap=self.energy_gap + field_energy_change,
        )

    def _verify_modes(self):
        """Verifies that the model has exactly two modes.

        Raises
        ------
        ValueError
            if the model does not have exactly two modes.
        """
        if len(self.mode_frequencies) != 2 or len(self.mode_couplings) != 2:
            raise ValueError("The MLJ model requires exactly two modes.")

    def __post_init__(self):
        self._verify_modes()
